package bsplines;

/*
Tutorial on LevelSet and BSpline deformable model techniques 
for image segmentation
Copyright (C) Diane Lingrand (Diane.Lingrand@unice.fr)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

import java.util.ArrayList;
import java.awt.Graphics;
import java.awt.Color;

/**
 * A BSpline is composed from several BSPoint organized in a ArrayList
 *
 * @author <a href="mailto:Diane.Lingrand@unice.fr">DianeLingrand</a>
 **/

public class BSpline {
    
    public static final int CIRCLE = 0;
    public static final int ELLIPSE = 1;
    public static final int TORDU = 2;

    protected static final int CLOSE = 10; 
    protected static final int DISTANT = 50;
    private ArrayList points;
   
    private int close = CLOSE;
    private int distant = DISTANT;

    private int surface;

    /** 
     * default constructor : constructs an empty BSpline
     **/
    public BSpline() {
	points = new ArrayList();
    }

    /** 
     * constructor : constructs a list of BSPoint arranged around a CIRCLE or an ELLIPSE or a TORDU form
     *
     * @param form CIRCLE or ELLIPSE or TORDU
     * @param w width of the image and the window
     * @param h height of the image and the window
     **/
    public BSpline(int form, int w, int h) {
	this();
	switch(form) {
	case CIRCLE:
	    int xCenter = w/2;
	    int yCenter = h/2;
	    int radius = (int) (0.45*Math.min(w,h));
	    int nbPoints = 40;
	    double theta = 2.0 * Math.PI / nbPoints;
	    double angle = 0.0;
	    for(int i = 0; i < nbPoints; i++) {
		BSPoint p = new BSPoint( xCenter + radius*Math.cos(angle), yCenter + radius*Math.sin(angle));
		points.add(p);
		angle += theta;
	    }
	    break;
	case ELLIPSE:
	    xCenter = w/2;
	    yCenter = h/2;
	    int aRadius = (int) (0.45 * w);
	    int bRadius = (int) (0.45 * h);
	    nbPoints = 40;
	    theta = 2.0 * Math.PI / nbPoints;
	    angle = 0.0;
	    for(int i = 0; i < nbPoints; i++) {
		BSPoint p = new BSPoint( xCenter + aRadius*Math.cos(angle), yCenter + bRadius*Math.sin(angle));
		points.add(p);
		angle += theta;
	    }
	    break;
	case TORDU:
	    xCenter = 2*w/3;
	    yCenter = h/2;
	    radius = (int) (0.20*Math.min(w,h));
	    nbPoints = 20;
	    theta = 2.0 * Math.PI / nbPoints;
	    angle = 0.0;
	    for(int i = 0; i < 10; i++) {
		BSPoint p = new BSPoint( xCenter + radius*Math.cos(angle), yCenter + radius*Math.sin(angle));
		points.add(p);
		angle += theta;
	    }
	    xCenter = w/3;
	    angle = 0.0;
	    for(int i = 0; i < 20; i++) {
		BSPoint p = new BSPoint( xCenter + radius*Math.cos(angle), yCenter + radius*Math.sin(angle));
		points.add(p);
		angle -= theta;
	    }
	    xCenter = 2*w/3;
	    angle = Math.PI;
	    for(int i = 0; i < 10; i++) {
		BSPoint p = new BSPoint( xCenter + radius*Math.cos(angle), yCenter + radius*Math.sin(angle));
		points.add(p);
		angle += theta;
	    }
	    break;
	default:break;
	}
	this.computeQ();
    }
    
    public int getSurface() { return surface;}
    public void setClose(int c) { close = c; }
    public void setDistant(int d) { distant = d;}

    /**
     * returns the number of P points (or Q points) of this BSpline
     **/
    public int size() {
	return points.size();
    }

    /**
     * computes and returns the length of the curve with "precision" epsilon
     * 
     * @param epsilon related to the precision of the evaluated length of the curve
     **/
    public double length(double epsilon) {
	int size = points.size();

	if(size < 2) return 0.0;

	double result = 0.0;

	// the total length is the sum of the length of each segment
	for(int i = 0; i < size; i++) {
	    BSPoint p = (BSPoint) points.get(i);
	    result += p.length(0,1,epsilon/size);
	}
	return result;
    }

    /**
     * creates and returns a mask for a rectangular region specified by the origin (x0,y) and the size (w,h)
     * the mask if a 2 values table where :
     *   - 0 means outside
     *   - col means inside
     * This is done assuming that the point (x0,y0) is outside a BSpline.
     * The algorithm is a two passes region coloring algorithm.
    **/
    public int [][] createMask(int [][] m, int x0, int y0, int w, int h, int col) {
	int [][] res = new int[w][h];
	int size = points.size();

	int aire = 0;
	
	// trivial case
	if(size < 2) 
	    return res;

	// first, we draw the curve with color = -1
	for(int i = 0; i < size; i++) {
	    BSPoint p = ((BSPoint) points.get(i));
	    p.trace(res,-1);
	}

	// then we mark with positive colors the different regions
	int color = 0;
	int wh = w*h;
	int [] corresp = new int[wh];
	for(int i = 0; i < wh; i++) 
	    corresp[i] = i;

	// we assume that the point [0][0] is outside the spline. then, the
	// color of this point is 0
	res[0][0] = color;
	
	//// FIRST PASS
	// case of the first line.
	for(int j = 1; j < h; j++)
	    if(res[0][j] !=-1){
		if(res[0][j-1] != -1)
		    res[0][j] = res[0][j-1];
		else
		    res[0][j] = ++color;
	    }

	// other lines i
	for(int i = 1; i < w; i++) {
	    // case of j=0
	    if(res[i][0] != -1) {
		if(res[i-1][0] != -1)
		    res[i][0] = res[i-1][0];
		else
		    res[i][0] = ++color;
	    }
	    // other columns j
	    for(int j = 1; j < h; j++) {
		if(res[i][j] !=-1) {
		    if(res[i-1][j] !=-1) {
			res[i][j] = res[i-1][j];
			if( (res[i][j-1] != -1) && (res[i][j] != res[i][j-1]) ) {
			    // two different colors correspond to the same region
			    int min;
			    int max;
			    if(res[i][j-1] < res[i][j]) {
				min = res[i][j-1];
				max = res[i][j];
			    }
			    else {
				min = res[i][j];
				max = res[i][j-1];
			    }
			    res[i][j] = min; 

			    SegPanel.tryCorresp(corresp,min,max);
			}
		    }
		    else if(res[i][j-1] != -1)
			res[i][j] = res[i][j-1];
		    else
			res[i][j] = ++color;
		}
	    }
	}
	//// UPDATING the table of color correspondances
	for(int i = 0; i < wh; i++) 
	    corresp[i] = corresp[corresp[i]];

	//// SECOND PASS
	for(int i = 0; i < w; i++)
	    for(int j = 0; j < h; j++) {
		// use the table of color correspondances
		if(res[i][j] != -1)
		    res[i][j] = corresp[res[i][j]];
		// 0 for outside, 1 for inside
		if(res[i][j] != 0) {
		    res[i][j] = col;
		    m[i][j] += col;
		    aire ++;
		}
	    }
	surface = aire;
	return res;
    }

    /**
    * computes the Q-points positions from the P-points positions and also
    * the a[] and b[] coefficients. 
    **/
 
    public void computeQ() {
	int size = points.size();
	if(size < 3) return;

	double [] xQ = new double[size];
	double [] yQ = new double[size];
	double [] xP = new double[size];
	double [] yP = new double[size];
	double [] a = new double[4];
	double [] b = new double[4];

 	for(int i = 0; i < size; i++) {
	    BSPoint p = ((BSPoint) points.get(i));
	    xP[i] = p.getXP();
	    yP[i] = p.getYP();
	}

	// initialisation of M matrix for which [Pi] = [M].[Qi]
	double [][] M = new double[size][size];
	for(int i = 0; i < size; i++)
	    for(int j = 0; j < size; j++)
		M[i][j] = 0.0;
	for(int i = 1; i < size - 1; i++) {
	    M[i][i-1] = 1.0;
	    M[i][i] = 4.0;
	    M[i][i+1] = 1.0;
	}
	M[0][0] = 4.0; M[0][1] = 1.0; M[0][size-1] = 1.0;
	M[size-1][0] = 1.0; M[size-1][size-2] = 1.0; M[size-1][size-1] = 4.0;

	// Gauss ...
	for(int i = 1; i < size; i++) {
	    double c = M[i-1][i-1];
	    xP[i] = c*xP[i] - xP[i-1];
	    yP[i] = c*yP[i] - yP[i-1];
	    for(int j = 0; j < size; j++) 
		M[i][j] = c*M[i][j] - M[i-1][j];
	}
	// except the last line, the M matrix is now triangular
	for(int i = 0; i < size - 1; i++) {
	    double c = M[size-1][i];
	    xP[size-1] -= xP[i]*c/M[i][i];
	    yP[size-1] -= yP[i]*c/M[i][i];
	    for(int j = 0; j < size; j++) 
		M[size-1][j] -= M[i][j]*c/M[i][i];
	}
	// the matrix M is now triangular (sup)

	// special case of size-1
	xQ[size-1] = 6.0 * xP[size-1]/M[size-1][size-1];
	yQ[size-1] = 6.0 * yP[size-1]/M[size-1][size-1];
	// special case of size-2
	xQ[size-2] = (6.0 * xP[size-2] - xQ[size-1]*M[size-2][size-1]) / M[size-2][size-2];
	yQ[size-2] = (6.0 * yP[size-2] - yQ[size-1]*M[size-2][size-1]) / M[size-2][size-2];
	// other cases
	for(int i = size-3 ; i >= 0; i--) {
	    xQ[i] = (6.0 * xP[i] - xQ[i+1]*M[i][i+1] - xQ[size-1]*M[i][size-1]) / M[i][i];
	    yQ[i] = (6.0 * yP[i] - yQ[i+1]*M[i][i+1] - yQ[size-1]*M[i][size-1]) / M[i][i];
	}

	for(int i = 0; i < size; i++) {
	    // setQ
	    BSPoint p = ((BSPoint) points.get(i));
	    p.setQ(xQ[i],yQ[i]);

	    // compute a[] and b[]
	    int im1 = (i-1+size)%size ;
	    int ip1 = (i+1)%size ;
	    int ip2 = (i+2)%size ;
	    a[0] = (xQ[im1] + 4*xQ[i] + xQ[ip1]) / 6.0;
	    b[0] = (yQ[im1] + 4*yQ[i] + yQ[ip1]) / 6.0;
	    a[1] = (xQ[ip1] - xQ[im1]) * 0.5;
	    b[1] = (yQ[ip1] - yQ[im1]) * 0.5;
	    a[2] = (xQ[ip1] + xQ[im1]) * 0.5 - xQ[i];
	    b[2] = (yQ[ip1] + yQ[im1]) * 0.5 - yQ[i];
	    a[3] = (xQ[ip2] - xQ[im1]) / 6.0 + (xQ[i] - xQ[ip1]) * 0.5;
	    b[3] = (yQ[ip2] - yQ[im1]) / 6.0 + (yQ[i] - yQ[ip1]) * 0.5;
	    
	    // set AB
	    p.setAB(a,b);
	}
    }

    /**
    * returns the i-th BSPoint
    **/
    public BSPoint get(int i) {
	if( (i<0) || (i>=points.size()) )
	    return null;
	return ((BSPoint) points.get(i));
    }

    /** 
     * adds a new BSPoint at the end of the list
     **/
    public void add(BSPoint p) {
	points.add(p);
    }
    /** 
     * adds a new BSPoint at the index idx
     **/
    public void add(int idx, BSPoint p) {
	points.add(idx, p);
    }

    /**
    * applies on each BSPoint (P-point) the constant force alphac 
    **/
    public void evolve(double alphac) {
	int size = points.size();
	BSPoint p;

	if(size <= 1) return;

	for(int i = 0; i < size; i++) {
	    p = ((BSPoint) points.get(i));
	    p.addP(100.0*p.getCurv()-alphac);
	}
	computeQ();
    }

    /**
    * removes points too close and adds points when too distant
    **/
    public void checkDistances() {
	BSPoint p;
	BSPoint pp1;

	if(points.size() == 1) {
	    points.remove(0);
	    return;
	}
	if(points.size() == 0) return;

	for(int i = 0; i < points.size() - 1 ; i++) {
	    p = ((BSPoint) points.get(i));
	    pp1 = ((BSPoint) points.get(i+1));
	    double dist = p.dist(pp1);
	    if(dist < close) {
		points.remove(i+1); System.out.println("removing " + (i+1) + " ..."); i--;}
	    else if( dist > distant) {
		BSPoint newP = p.getMiP();
		if(p.dist(newP) >= close) {
		    points.add(i+1, newP); System.out.println("adding " + (i+1) + " ..."); i++;}
	    }
	}
	if(points.size() == 1)  {
	    points.remove(0);
	    return;
	}
	if(points.size() == 0) return;

	p = ((BSPoint) points.get(points.size() - 1));
	pp1 = ((BSPoint) points.get(0));
	if(p.dist(pp1) < close) {
	    points.remove(0); System.out.println("removing 0 ...");}
	
	computeQ();
    }

    /** 
     * computes the length of this BSpline et re-organizes P-points along this curve.
     * We put the new points on the previous curve with the constraint that the curve
     * length between two consecutives points is a constant, determined as the previous
     * length divided by the previous number of points. This constant must also be greater
     * than "close" and smaller than "distant"
     *
     * @param var if equal to zero, nothing different. If > 0, we add points. If < 0, we delete points,
     *            according to the "close" and "distant" limits.
     **/
    public void regulDistances(int var) {

	// trivial cases 
	if(points.size() == 1) {
	    points.remove(0);
	    return;
	}
	if(points.size() == 0) return;

	int size = points.size();
	double epsilon = 0.0000000001;
	double presqueUn = 0.9999999999;
	double length = length(epsilon);

	// too short BSpline => erase
	if(length < 2*close) {
	    points = new ArrayList();
	    return;
	}
	// TDB ajouter egalement une condition entre l'aire et le perimetre pour eliminer
	// les regions trop fines
	if(size + var == 0)
	    return;
	double dist = length / (size + var);
	dist = (dist > close) ? dist : close;
	dist = (dist > distant) ? distant : dist;
	int nbPoints = (int) (length / dist);
	
	// creation of a new list of P-points with the first point P0
	ArrayList newList = new ArrayList();
	BSPoint p = (BSPoint) points.get(0);
	newList.add(p);
	nbPoints --;

	// travel on the curve
	BSPoint pCurr = p; // first point of the current segment
	                   // p is the current point
	int iNext = 0;       // index of the current segment
	double t = 0.0;    // t value of the current point (p)
	// p.findT(debut, value, epsilon) enables to find a t value for which : |arc_length(P(debut),P(t))-value|<epsilon
	// p.pointAtTvalue(t) gives P(t)

	boolean stop = false;
	boolean firstSeg = true;

	while((!stop) && (nbPoints > 0)) {
	// recherche du point suivant
	// recherche du segment du point suivant
	    
	    // est-il dans le segment courant ?
	    if(dist < pCurr.length(t,1.0,epsilon)) {
		// oui : eh bien on l'ajoute
		t = pCurr.findT(t, dist, epsilon);
		p = pCurr.pointAtTvalue(t);

		newList.add(p);
		nbPoints--;
	    }
	    else {
		// non => recherche dans les autres segments
		
		// on cherche tout d'abord le numero de segment dans lequel rechercher notre point
		iNext++;
		firstSeg = false;
		if(iNext == size) iNext = 0;
		double value = dist - pCurr.length(t,1.0,epsilon);
		pCurr = (BSPoint) points.get(iNext);
		double newLength =  pCurr.length(epsilon);

		while( (iNext!=0) && (value > newLength )) {
		    value -= newLength;
		    iNext++; 
		    if(iNext == size) iNext = 0;
		    pCurr = (BSPoint) points.get(iNext);
		    newLength = pCurr.length(epsilon);
		}

		// le segment ou trouver le point est iNext (sauf si t=1 alors iNext+1)
		if(iNext == 0) stop = true;
		else {
		    // ajout d'un point du segment iNext commencant par pCurr
		    t = pCurr.findT(0.0,value,epsilon);

		    p = pCurr.pointAtTvalue(t);
		    newList.add(p);
		    nbPoints --;
		}
	    }
	}       

	// ending
	points = newList;
	computeQ();

	size = newList.size();

	// if the last segment (between size-1 and 0) is too big, we divide it by 2
	BSPoint p0 = (BSPoint) points.get(0);
	BSPoint pn = (BSPoint) points.get(size-1);
	if(pn.length(epsilon) > 1.5*dist)
	    points.add(pn.pointAtTvalue(pn.findT(0.0,0.75*dist,epsilon)));

	computeQ();
	size = newList.size();
    }

    /** 
     * clones this BSpline
     **/
    public java.lang.Object clone() {
	BSpline bsClone = new BSpline();
	int size = this.size();

	for(int i = 0; i < size; i++) {
	    BSPoint p = (BSPoint) ((BSPoint) points.get(i)).clone();
	    bsClone.add(p);
	}

	return bsClone;
    }

    /** 
     * test if there is an auto-intersection of this BSpline.
     * The test is done using a window (w x h) in which we draw the BSpline.
     **/
    public boolean isIntersect(int w, int h) {
	int inter = -1;
	int [][] table = new int[w][h];
	int i = 0;
	int size = points.size();

	if(size < 3)
	    return false;
	int x0 = (int) ( ((BSPoint) points.get(0)).getXP() + 0.5);
	int y0 = (int) ( ((BSPoint) points.get(0)).getYP() + 0.5);
	while( (inter == -1) && (i < size)) {
	    inter = ((BSPoint) points.get(i)).traceB(table, i+1,x0,y0);
	    i++;
	}
	if(inter == -1)
	    return false;
	else
	    return true;
    }

    /** 
     * test if there is an auto-intersection of this BSpline.
     * if not, returns null
     * if yes, cut this BSpline in two parts, one for this BSpline and the other returned
     **/
    public BSpline cutBSpline(int w, int h) {
	BSpline newBS = null;
	ArrayList currBS = new ArrayList();
	int size = points.size();
	int i = 0;
	int inter = -1;
	int [][] table = new int[w][h];

	// trivial case
	if(size < 3) return null;

	BSPoint p = (BSPoint) points.get(0); //current point
	int x0 = (int) (0.5 + p.getXP());
	int y0 = (int) (0.5 + p.getYP());
	// search the segment of intersection
	while( (inter==-1) && ( i < size) )  {
	    p = (BSPoint) points.get(i);
	    inter = p.traceB(table, i+1, x0, y0);
	    //if(inter != -1) System.out.println("inter = " + inter + " et i = " + i);
	    i++;
	}

	if( inter != -1 ) {
	    // there is intersection in the (i-1)th segment beginning with p with the segment (inter-1)
	    // we build now 2 splines
	    // one from 0 to inter-1 and i to size-1
	    // the other from inter to i-1

	    for(int j = 0; j < inter; j++) {
		p = (BSPoint) points.get(j);
		currBS.add(p);
	    }
	    newBS = new BSpline();

	    for(int j = i; j < size; j++)
		currBS.add((BSPoint) points.get(j));

	    for(int j = inter; j < i; j++) {
		p = (BSPoint) points.get(j);
		newBS.add(0,p);
	    }

	    points = currBS;
	    if(size() >= 3)
		computeQ();

	    if(newBS.size() >= 3)
		newBS.computeQ();

	    System.out.print("currBS with " + currBS.size() + " points");
	}
	    
	if(newBS !=null)
	    System.out.println(" and newBS with " + newBS.size() +" points");

	return newBS;
    }

    /**
    * draws the BSpline in a Graphics g
    **/
    public void paint(Graphics g, Color c) {
	int size = points.size();
	//	System.out.println("size = " + size);
	for(int i = 0; i < size; i++)
	    ((BSPoint) points.get(i)).paint(g,c);
    }

    /**
     * String formatting of this object
     **/
    public String toString() {
	int size = points.size();
	String result = "BSpline de " + size + " points.\n";
	for(int i = 0; i < size; i++)
	    result += ((BSPoint) points.get(i)).toString() +"\n";
	return result;
    }

    /** 
     * the same as toString but with another formatting, used in BSPanel.saveBSplines
     **/
    public String createString() {
	int size = points.size();
	String result = "" + size + "\n";
	for(int i = 0; i < size; i++) {
	    BSPoint p = ((BSPoint) points.get(i));
	    result += p.getXP() + " " + p.getYP()+"\n";;
	}
	return result;
    }

    /**
     * check if the orientation is OK (normals looking inside the object)
     * the simple algorithm is based that, often (not always), if we take 3 points P0, P1 and P2
     * along the curve, verifying that the arc distances P0P1, P1P2 and P2P0 are equals,
     * the sign of the cross product P0P1xP0p2 must be positive. If not, we inverse the points
     * order in the BSpline
     **/
    public void reorientCurv() {
	if(!isOrientationOK()) {
	    ArrayList newList = new ArrayList();
	    int size = points.size();
	    for(int i = size - 1 ; i >= 0 ; i--)
		newList.add((BSPoint) points.get(i));
	    points = newList;
	    computeQ(); // we need to recomputes the normal vectors and cie ...
	}
    }

    public boolean isOrientationOK() {
	int size = points.size();
	BSPoint p0, p1, p2;
	p0 = (BSPoint) points.get(0);
	p1 = (BSPoint) points.get(size/3);
	p2 = (BSPoint) points.get(2*size/3);
	double crossProduct = (p1.getXP()-p0.getXP())*(p2.getYP()-p0.getYP()) - 
	    (p1.getYP()-p0.getYP())*(p2.getXP()-p0.getXP());
	if(crossProduct > 0)
	    return true;
	return false;
    }
	
    /**
    * main ... for testing ...
    **/
    public static void main(String [] args) {
	BSpline bsp = new BSpline();
	
    }
}
