package bsplines;

/*
Tutorial on LevelSet and BSpline deformable model techniques 
for image segmentation
Copyright (C) Diane Lingrand (Diane.Lingrand@unice.fr)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

import javax.swing.JPanel;
import javax.swing.*;
import javax.imageio.*;
import java.io.*;
import java.awt.*;
import java.awt.image.*;
import java.net.URL;

public abstract class SegPanel extends JPanel {
    protected static final String IMAGE_FILENAME="test.jpg";
    protected static final int MEAN = 0, VARIANCE1 = 1, VARIANCE2 = 2, GRADIENT1 = 3, GRADIENT2 = 4, GRADIENT3 = 5, GRADIENT4 = 6;
    protected static final Color [] BS_COLORS = { new Color(0,255,0),  new Color(0,200,255), new Color(150,255,40),  new Color(0,40,255),  new Color(0,255,80), new Color(250,200,255), new Color(0,160,255),  new Color(0,255,120), new Color(160,200,255), new Color(0,80,255), new Color(0,255,160), new Color(0,120,255), new Color(0,255,200), new Color(0,255,255), new Color(0,0,255), new Color(80,200,255)};
    
    protected double lambda1, lambda2, lambda3;
    protected int criterion = MEAN;
    protected BufferedImage bi = null;
    protected BufferedImage biLiss = null;
    protected int w = 200;
    protected int h = 200;

    protected boolean avecLissage = false;
    protected boolean capture = false; // do we save an image file at each iteration ?

    public abstract void clearBSplines();
    public abstract void evolve(boolean b);
    public abstract void regulDistances(int var);
    public abstract void checkOverlapping();
    public abstract void cutBSplines();
    public abstract void reorientCurv();
    public abstract void loadBSplines(BufferedReader reader) ;
    public abstract void saveBSplines(File file) ;
    public abstract String saveBSplinesAsString() ;
    public abstract BSparam loadParameters(BufferedReader reader) ;
    public abstract void saveParameters(File file) ;
    public abstract void saveParameters(JTextArea jt) ;
    public abstract void setDistant(int d);
    public abstract void setClose(int c);
    public abstract void distMapShow();
    public abstract void evolveMean();
    public abstract void evolveVariance1();
    public abstract void evolveVariance2();
    public abstract void evolveGradient1();
    public abstract void evolveGradient2();
    public abstract void evolveGradient3();
    public abstract void evolveGradient4();

    public void setCapture(boolean b) {capture = b;}
    public boolean getCapture() { return capture;}

    public void setSmoothing(boolean b) {avecLissage = b;}
    public boolean getSmoothing() { return avecLissage;}

    public void setCriterion(int c) {  criterion = c;    }
    public void setLambda1(double l) { lambda1 = l;}
    public void setLambda2(double l) { lambda2 = l;}
    public void setLambda3(double l) { lambda3 = l;}
    protected int getIavtLiss(int x, int y) {	
	int red = (bi.getRGB(x, y)>>16)& 0x000000ff ; 
	int green = (bi.getRGB(x, y)>>8)& 0x000000ff ; 
	int blue = bi.getRGB(x, y)& 0x000000ff ; 
	int col;

	// used for the hand
	//if(green+blue >= 1) col = (int) (255*(0.5 + red*1.0/(green+blue)));
	//else col = red;

	//assuming the image is in grey level (r=g=b)
	col = red;
	return col;
    }
    protected int getI(int x, int y) {
	if(avecLissage)
	    return biLiss.getRGB(x, y)& 0x000000ff ; 
	
	return getIavtLiss(x,y); 
    }
    protected BufferedImage getBufferedImage() { return bi;}
    protected BufferedImage getLissBufferedImage() { return biLiss;}
    protected double getLambda1() {return lambda1;}
    protected double getLambda2() {return lambda2;}
    protected double getLambda3() {return lambda3;}
    
    public void saveImage(File file) {
	try {
	    w = bi.getWidth();
	    h = bi.getHeight();
	    BufferedImage toSave = new BufferedImage(w,h,BufferedImage.TYPE_INT_RGB);
	    paintComponent(toSave.getGraphics());
	    ImageIO.write(toSave,"png",file);
	}
	catch ( IOException e ) {
	    System.out.println("Unable to save in file " + file);
	}
    }

    /**
     * changes the current image (and keeps the other parameters as before)
     **/
    public void changeImage(File file) {
	try {
	    bi = ImageIO.read(file);
	    w = bi.getWidth();
	    h = bi.getHeight();
	    setPreferredSize(new Dimension(w,h));
	    biLiss = lissage(bi);
	}
	catch ( IOException e ) {
	    System.out.println("Unable to open file " + file);
	}
	repaint();
    }

    public void changeImage(URL url) {
	ImageIcon ii = null;
	try {
	    ii = new ImageIcon(url);
	}
	catch(Exception e) {
	    JOptionPane.showMessageDialog(this, "It is not allowed to access this file","Error", JOptionPane.ERROR_MESSAGE);
	}
	int wi = ii.getIconWidth();
	int hi = ii.getIconHeight();
	if( (wi > 0) && ( hi > 0) ) {
	    bi = new BufferedImage(wi, hi, BufferedImage.TYPE_INT_RGB);
	    Graphics g = bi.getGraphics();
	    ii.paintIcon(this, g, 0, 0);
	    w = wi;
	    h = hi;
	    biLiss = lissage(bi);
	    repaint();	
	}
    }

    public BufferedImage lissage(BufferedImage b) {
	if(b == null) return null;
	int w = b.getWidth();
	int h = b.getHeight();

	BufferedImage res = new BufferedImage(w,h,BufferedImage.TYPE_INT_RGB);
	int col;
	for(int i = 1; i < w-1 ; i ++) 
	    for(int j = 1; j < h-1; j++) {
		// mean with neighbours
		col = (getIavtLiss(i+1,j-1) + 2*getIavtLiss(i+1,j)+ getIavtLiss(i+1,j+1)+ 2*getIavtLiss(i,j-1)+ 4*getIavtLiss(i,j)+ 2*getIavtLiss(i,j+1)+ 2*getIavtLiss(i-1,j)+ getIavtLiss(i-1,j-1)+ getIavtLiss(i-1,j+1))/16;
		res.setRGB(i,j,col*65793);
	    }
	// borders
	for(int i = 0; i < w ; i ++) {
	    res.setRGB(i,0,getIavtLiss(i,0)*65793);
	    res.setRGB(i,h-1,getIavtLiss(i,h-1)*65793);
	}
	for(int j = 1; j < h-1; j++) {
	    res.setRGB(0,j,getIavtLiss(0,j)*65793);
	    res.setRGB(w-1,j,getIavtLiss(w-1,j)*65793);
	}
	return res;
    }
		
    /**
     *  applies the forces related to the choosed criterion on 
     * each BSPoint of each BSpline of the list of BSplines
     **/
    public void applyCrit() {
	switch(criterion) {
	case MEAN:	    evolveMean();	    break;
	case VARIANCE1:	    evolveVariance1();	    break;
	case VARIANCE2:	    evolveVariance2();	    break;
	case GRADIENT1:	    evolveGradient1();	    break;
	case GRADIENT2:	    evolveGradient2();	    break;
	case GRADIENT3:	    evolveGradient3();	    break;
	case GRADIENT4:	    evolveGradient4();	    break;
	default:	    evolveMean();	    break;
	}
    }

    /** 
     * used to managed color correspondances 
     * taking into account the fact that we want
     * to replace 6->5 and 6->0 by 5->0 and 6->0
     **/
    protected static void tryCorresp(int [] co, int min, int max) {
	// already done
	if(co[max] == min)
	    return;
	// easy case
	if(co[max] == max) {
	    co[max] = min;
	    return;
	}
	// other cases
	int exco = co[max];
	if(exco < min) {
	    tryCorresp(co,exco,min);
	}
	else if(exco > min) {
	    tryCorresp(co,min,exco);
	    co[max]=min;
	}
    }

}
